// *********************
// * NoHeadShot Plugin *
// *********************
// version 1.12
// by Pierre-Marie Baty <pm@bots-united.com>
// and AustinBotServer <austinbotserver@mindspring.com>

// LAST MODIFIED BY PM (change this line when you modify the CPP file)

#include <extdll.h>
#include <dllapi.h>
#include <h_export.h>
#include <meta_api.h>


// parts of body for hits
#define HITGROUP_GENERIC 0 // none
#define HITGROUP_HEAD 1
#define HITGROUP_CHEST 2
#define HITGROUP_STOMACH 3
#define HITGROUP_LEFTARM 4
#define HITGROUP_RIGHTARM 5
#define HITGROUP_LEFTLEG 6
#define HITGROUP_RIGHTLEG 7   


enginefuncs_t g_engfuncs;
globalvars_t  *gpGlobals;
DLL_FUNCTIONS gFunctionTable;

void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals);
void TraceLine_Post (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr);


// START of Metamod stuff

enginefuncs_t meta_engfuncs_Post;
gamedll_funcs_t *gpGamedllFuncs;
mutil_funcs_t *gpMetaUtilFuncs;
meta_globals_t *gpMetaGlobals;

META_FUNCTIONS gMetaFunctionTable =
{
   NULL, // pfnGetEntityAPI()
   NULL, // pfnGetEntityAPI_Post()
   NULL, // pfnGetEntityAPI2()
   NULL, // pfnGetEntityAPI2_Post()
   NULL, // pfnGetNewDLLFunctions()
   NULL, // pfnGetNewDLLFunctions_Post()
   NULL, // pfnGetEngineFunctions()
   GetEngineFunctions_Post, // pfnGetEngineFunctions_Post()
};

plugin_info_t Plugin_info = {
   META_INTERFACE_VERSION, // interface version
   "NoHeadShot", // plugin name
   "1.12", // plugin version
   "22/05/2003", // date of creation
   "Pierre-Marie Baty <pm@bots-united.com> & AustinBotServer <austinbotserver@mindspring.com>", // plugin author
   "http://racc.bots-united.com", // plugin URL
   "NOHS", // plugin logtag
   PT_CHANGELEVEL, // when loadable
   PT_ANYTIME, // when unloadable
};


C_DLLEXPORT int Meta_Query (char *ifvers, plugin_info_t **pPlugInfo, mutil_funcs_t *pMetaUtilFuncs)
{
   // this function is the first function ever called by metamod in the plugin DLL. Its purpose
   // is for metamod to retrieve basic information about the plugin, such as its meta-interface
   // version, for ensuring compatibility with the current version of the running metamod.

   // keep track of the pointers to metamod function tables metamod gives us
   gpMetaUtilFuncs = pMetaUtilFuncs;
   *pPlugInfo = &Plugin_info;

   // check for interface version compatibility
   if (strcmp (ifvers, Plugin_info.ifvers) != 0)
   {
      int mmajor = 0, mminor = 0, pmajor = 0, pminor = 0;

      LOG_CONSOLE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);
      LOG_MESSAGE (PLID, "%s: meta-interface version mismatch (metamod: %s, %s: %s)", Plugin_info.name, ifvers, Plugin_info.name, Plugin_info.ifvers);

      // if plugin has later interface version, it's incompatible (update metamod)
      sscanf (ifvers, "%d:%d", &mmajor, &mminor);
      sscanf (META_INTERFACE_VERSION, "%d:%d", &pmajor, &pminor);

      if ((pmajor > mmajor) || ((pmajor == mmajor) && (pminor > mminor)))
      {
         LOG_CONSOLE (PLID, "metamod version is too old for this plugin; update metamod");
         LOG_ERROR (PLID, "metamod version is too old for this plugin; update metamod");
         return (FALSE);
      }

      // if plugin has older major interface version, it's incompatible (update plugin)
      else if (pmajor < mmajor)
      {
         LOG_CONSOLE (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         LOG_ERROR (PLID, "metamod version is incompatible with this plugin; please find a newer version of this plugin");
         return (FALSE);
      }
   }

   return (TRUE); // tell metamod this plugin looks safe
}


C_DLLEXPORT int Meta_Attach (PLUG_LOADTIME now, META_FUNCTIONS *pFunctionTable, meta_globals_t *pMGlobals, gamedll_funcs_t *pGamedllFuncs)
{
   // this function is called when metamod attempts to load the plugin. Since it's the place
   // where we can tell if the plugin will be allowed to run or not, we wait until here to make
   // our initialization stuff, like registering CVARs and dedicated server commands.

   // are we allowed to load this plugin now ?
   if (now > Plugin_info.loadable)
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT attaching (can't load plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from attaching this plugin
   }

   // keep track of the pointers to engine function tables metamod gives us
   gpMetaGlobals = pMGlobals;
   memcpy (pFunctionTable, &gMetaFunctionTable, sizeof (META_FUNCTIONS));
   gpGamedllFuncs = pGamedllFuncs;

   // print a message to notify about plugin attaching
   LOG_CONSOLE (PLID, "%s: plugin attaching", Plugin_info.name);
   LOG_MESSAGE (PLID, "%s: plugin attaching", Plugin_info.name);

   return (TRUE); // returning TRUE enables metamod to attach this plugin
}


C_DLLEXPORT int Meta_Detach (PLUG_LOADTIME now, PL_UNLOAD_REASON reason)
{
   // this function is called when metamod unloads the plugin. A basic check is made in order
   // to prevent unloading the plugin if its processing should not be interrupted.

   // is metamod allowed to unload the plugin ?
   if ((now > Plugin_info.unloadable) && (reason != PNL_CMD_FORCED))
   {
      LOG_CONSOLE (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      LOG_ERROR (PLID, "%s: plugin NOT detaching (can't unload plugin right now)", Plugin_info.name);
      return (FALSE); // returning FALSE prevents metamod from unloading this plugin
   }

   return (TRUE); // returning TRUE enables metamod to unload this plugin
}


C_DLLEXPORT int GetEngineFunctions_Post (enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion)
{
   // this is one of the initialization functions hooked by metamod in the engine API

   // keep track of pointers to the functions we use in the engine API
   meta_engfuncs_Post.pfnTraceLine = TraceLine_Post;

   // copy the whole table for metamod to know which functions we are using here
   memcpy (pengfuncsFromEngine, &meta_engfuncs_Post, sizeof (enginefuncs_t));
   return (TRUE); // alright
}

// END of Metamod stuff




#ifdef _WIN32
int WINAPI DllMain (HINSTANCE hinstDLL, unsigned long fdwReason, LPVOID lpvReserved)
{
   // required Win32 DLL entry point. Good place for freeing stuff upon DLL unloading (in case
   // fdwReason would be DLL_PROCESS_DETACH) if needed. The same for Linux would be implemented
   // in an optional _fini() function.

   return (TRUE); // nothing to do here
}
#endif


void WINAPI GiveFnptrsToDll (enginefuncs_t* pengfuncsFromEngine, globalvars_t *pGlobals)
{
   // this function is called by the engine to tell the MOD DLL (hence the plugins too) the
   // address of its function table and its server global variables structure.

   // keep track of the engine functions and the server globals for our own use
   memcpy (&g_engfuncs, pengfuncsFromEngine, sizeof (enginefuncs_t));
   gpGlobals = pGlobals;
}


void TraceLine_Post (const float *v1, const float *v2, int fNoMonsters, edict_t *pentToSkip, TraceResult *ptr)
{
   // this function is called right after the TraceLine() function is called, which traces a
   // line dot by dot, starting from vecStart in the direction of vecEnd, ignoring or not
   // monsters (depending on the value of igmon, ignore_monsters or dont_ignore_monsters), and
   // stops at the first obstacle encountered, returning the results of the trace in the
   // TraceResult structure ptr. Such results are (amongst others) the distance traced, the hit
   // surface, the hit plane vector normal, etc. See the TraceResult structure for details. This
   // function allows to specify whether the trace starts "inside" an entity's polygonal model,
   // and if so, to specify that entity in pentIgnore in order to ignore it as a possible
   // obstacle. Hence by hooking it we can know, by looking in the resulting TraceResult
   // structure, if the entity that is hit is a human, if he was hit in the head, and if the
   // entity that "fired" the traceline was a bot. In which case, the conditions are met for
   // saying safely that it's a bot that fired its gun and got a headshot to a player. Since
   // we don't want this to happen, we can modify the resulting TraceResult structure so that
   // it fakes as if the player wasn't hit at all.

   // was this TraceLine fired by an entity, AND is this entity firing the traceline a bot,
   // AND is the entity that was hit a human player, AND was he hit in the head ?
   if (!FNullEnt (pentToSkip) && (pentToSkip->v.flags & FL_FAKECLIENT)
       && !(ptr->pHit->v.flags & FL_FAKECLIENT) && ((1 << HITGROUP_HEAD) & (1 << ptr->iHitgroup)))
      ptr->flFraction = 1.0; // we don't want that, so hack the trace results to discard the hit

   RETURN_META (MRES_HANDLED); // now return the new (modified) TraceResult structure
}
